#ifndef __LIBEINA_CORE_H__
#define __LIBEINA_CORE_H__

#include <glib.h>
#include <gst/gst.h>
#include <gst/control/control.h>
#include <libeina/config.h>
#include <libeina/stream.h>

G_BEGIN_DECLS

typedef struct _EinaCore EinaCore;

/* Map some types from GStreamer, for naming convenience */
typedef enum {
	EINA_CORE_VOID  = GST_STATE_VOID_PENDING,
	EINA_CORE_NULL  = GST_STATE_NULL,
	EINA_CORE_STOP  = EINA_CORE_NULL, 
	EINA_CORE_READY = GST_STATE_READY,
	EINA_CORE_PAUSE = GST_STATE_PAUSED,
	EINA_CORE_PLAY  = GST_STATE_PLAYING
} EinaCoreState;

/* Idem for states */
typedef enum {
	EINA_CORE_FAILURE  = GST_STATE_FAILURE,
	EINA_CORE_SUCCESS  = GST_STATE_SUCCESS,
	EINA_CORE_ASYNC    = GST_STATE_ASYNC
} EinaCoreStateReturn;

/* Some error codes */
typedef enum {
	EINA_CORE_ERROR_NO_PIPELINE,
	EINA_CORE_ERROR_NO_SRC,
	EINA_CORE_ERROR_NO_SPIDER,
	EINA_CORE_ERROR_NO_SINK,
	EINA_CORE_ERROR_NO_VOLUME,
	EINA_CORE_ERROR_NO_VOLUME_DPARAM,
	EINA_CORE_ERROR_NO_MUTE_DPARAM
} EinaCoreError;

/*
 * @eina_core_init
 * @ Initializes Gstreamer library
 *  gint* argc: Number of parameters
 *  gchar** argv[]: A null ended list of arguments
 */
void eina_core_init(gint *argc, gchar **argv[]);

/*
 * @eina_core_new
 * @ Creates a new EinaCore object
 *  EinaCore* ret: A new EinaCore object
 */
EinaCore *eina_core_new(GError **err); 

/*
 * @eina_core_unref
 * @ Deletes a EinaCore stream
 *  EinaCore *self: EinaCore object to delete 
 */
void eina_core_unref(EinaCore *self);

/*
 * @eina_core_signal_connect
 * @ Connects underlaying events of GStreamer pipeline to functions
 *  EinaCore* p: [self]
 *  gchar* signal: Event to connect
 *  GCallback function: Function to call when event `signal` happends
 *  gpointer data: extra data to pass to callback function
 */
void eina_core_signal_connect(EinaCore *p, gchar *signal, GCallback function, gpointer data);

/*
 * @eina_core_set_input
 * @ Sets input plugin for reading streams
 *  EinaCore* p: [self]
 *  gchar* driver: New gstreamer input plugin to use
 *  GError** err: Human readable error status
 *  gboolean ret: Status of operation
 */
gboolean eina_core_set_input (EinaCore *p, gchar* driver, GError **err);

/*
 * @eina_core_set_output
 * @ Sets output plugin to use
 *  EinaCore* p: [self]
 *  gchar* driver: New gstreamer output plugin to use
 *  GError** err: Human readable error status
 *  gboolean ret: Status of operation
 */
gboolean eina_core_set_output(EinaCore *p, gchar* driver, GError **err);

/*
 * @eina_core_set_stream
 * @ Sets the stream to handle by the core
 *  EinaCore* p: [self]
 *  const EinaStream* stream: New stream to insert into core
 *  EinaCoreStateReturn ret: Exit status of operation
 */
EinaCoreStateReturn eina_core_set_stream(EinaCore *p, const EinaStream *stream);

/*
 * @eina_core_get_stream
 * @ Retrives the EinaStream that is currently inside the core
 *  EinaCore* p: [self]
 *  const EinaStream ret: 
 */
const EinaStream *eina_core_get_stream(EinaCore *p);

/*
 * @eina_core_get_state
 * @ Gets the current state of the core (playing, pause, stop, etc)
 *  EinaCore* p: [self]
 *  EinaCoreState ret: Current state of the core
 */
EinaCoreState eina_core_get_state(EinaCore *p);

/*
 * @eina_core_set_state
 * @ Sets the state of the core (playing, pause, stop, etc)
 *  EinaCore* p: [self]
 *  EinaCoreStateReturn ret: Exit code of the operation
 */
EinaCoreStateReturn eina_core_set_state(EinaCore *p, EinaCoreState state);

/*
 * @eina_core_set_volume
 * @ Sets the volume of core, value must be beetwen 0 and 100
 * @ lower and upper values are fixed to limits
 *  EinaCore* p: [self]
 *  gint vol: Value for volume
 */
gboolean eina_core_set_volume(EinaCore *p, gint vol);

/*
 * @eina_core_get_volume
 * @ Gets the current volume of the core
 *  EinaCore* p: [self]
 *  gint ret: The volume of the core
 */
gint eina_core_get_volume (EinaCore *p);

/*
 * @eina_core_set_mute
 * @ Sets or unsets the mute state on the core
 *  EinaCore* p: [self]
 *  gboolean mute: TRUE for mute the core, FALSE to use current volume
 */
gboolean eina_core_set_mute(EinaCore *p, gboolean mute);

/*
 * @eina_core_get_mute
 * @ Returns current state of mute state
 *  EinaCore* p: [self]
 *  gboolean ret: Current state of mute
 */
gboolean eina_core_get_mute (EinaCore *p);

/*
 * @eina_core_tell
 * @ Returns current position in microsecs of the current stream
 *  EinaCore* p: [self]
 *  gint64 ret: Current position in microsecs
 */
gint64 eina_core_get_position (EinaCore *p);

/*
 * @eina_core_seek
 * @ Moves position inside the stream, allowing seek.
 *  EinaCore* p: [self]
 *  gint64 pos: Position in microsecs where you want to move
 *  gboolean ret: Return status of operation
 */
gboolean eina_core_set_position(EinaCore *p, gint64 pos);

/*
 * @eina_core_length
 * @ Tells the lenght of stream in microsecs
 *  EinaCore* p: [self]
 *  gint64 ret: Lenght of stream in microsecs
 */
gint64 eina_core_length(EinaCore *p);

G_END_DECLS

#endif /* __LIBEINA_CORE_H__ */
