#ifndef __LIBEINA_PLAYLIST_H__
#define __LIBEINA_PLAYLIST_H__

#include <glib.h>
#include <libeina/stream.h>

G_BEGIN_DECLS

typedef struct _EinaPlaylist EinaPlaylist;

/* 
 * @eina_playlist_new
 * @ Creates a new playlist
 *  EinaPlaylist* ret: The new EinaPlaylist pseudo-object
 */
EinaPlaylist *eina_playlist_new(void);

/*
 * @eina_playlist_unref
 * @ Frees all resources from a EinaPlaylist object
 *  EinaPlaylist* l: [self]
 */
void eina_playlist_unref(EinaPlaylist *l);

/* 
 * @eina_playlist_add
 * @ Adds a EinaStream to playlist
 *  EinaPlaylist* l: [self]
 *  EinaStream* stream: Stream to add, it cannot be freed without previous eina_playlist_del call
 *  gboolean ret: TRUE on succesfull, FALSE on failure
 */
gboolean eina_playlist_add(EinaPlaylist *l, const EinaStream *stream);

/*
 * @eina_player_del
 * @ Deletes the EinaStream at position 'pos' from playlist
 *  EinaPlaylist* l: [self]
 *  guint pos: Position of the stream to delete.
 */
void eina_playlist_del(EinaPlaylist * l, guint pos);

/*
 * @eina_playlist_clear
 * @ Deletes all streams from playlist, this is not a deep free, EinaStream
 * @ contained in playlist sould be freed manualy
 *  EinaPlaylist* l: [self]
 */
void eina_playlist_clear(EinaPlaylist *l);

/*
 * @eina_playlist_get_playlist
 * @ Returns a GList of EinaStreams contained in playlist
 *  EinaPlaylist* l: [self]
 *  GList* ret: A GList of EinaStreams
 */
GList * eina_playlist_get_playlist(EinaPlaylist *l);

/*
 * @eina_playlist_get_nth
 * @Returns the EinaStream at position 'pos'
 *  EinaPlaylist* l: [self]
 *  guint pos: Postion to query
 *  const EinaStream* ret: EinaStream located at position 'pos'
 */
const EinaStream *eina_playlist_get_nth(EinaPlaylist *l, guint pos);

/*
 * @eina_playlist_total
 * @Gets the number of streams currently in playlist
 *  EinaPlaylist* l: [self]
 *  gint ret: The number of streams in playlist
 */
gint eina_playlist_total (EinaPlaylist *l);

/*
 * @eina_playlist_get_current
 * @Gets active stream in playlist, returns -1 if thereis no active element
 *  EinaPlaylist* l: [self]
 *  gint ret: Active stream
 */
gint eina_playlist_get_current(EinaPlaylist *l);

/* @eina_playlist_set_current
 * @Sets active stream in playlist. -1 sets no active element
 *  EinaPlaylist* l: [self]
 *  gint pos: Number of stream to set as active
 */
void eina_playlist_set_current(EinaPlaylist *l, gint pos);

/* 
 * @eina_playlist_get_random
 * @Returns if random mode is active or not
 *  EinaPlaylist* l: [self]
 *  gboolean ret: TRUE is random is active, otherwise returns FALSE
 */ 
gboolean eina_playlist_get_random(EinaPlaylist *l);

/* 
 * @eina_playlist_set_random
 * @Sets or unsets random mode
 *  EinaPlaylist* l: [self]
 *  gboolean val: TRUE for set random on, FALSE for off
 */
void eina_playlist_set_random(EinaPlaylist *l, gboolean val);

/* 
 * @eina_playlist_get_repeat
 * @Returns if repeat mode is active or not
 *  EinaPlaylist* l: [self]
 *  gboolean ret: TRUE is repeat is active, otherwise returns FALSE
 */ 
gboolean eina_playlist_get_repeat (EinaPlaylist *l);

/* 
 * @eina_playlist_set_repeat
 * @Sets or unsets repeat mode
 *  EinaPlaylist* l: [self]
 *  gboolean val: TRUE for set repeat on, FALSE for off
 */
void eina_playlist_set_repeat(EinaPlaylist *l, gboolean val);

/*
 * @eina_playlist_get_prev
 * @Returns the previous postion to active stream, if there is no previous position
 * @retuns -1
 *  EinaPlaylist* l: [self]
 *  gint ret: Previous position to active stream
 */
gint eina_playlist_get_prev(EinaPlaylist *l);

/*
 * @eina_playlist_get_next
 * @Returns the next postion to active stream, if there is no next position
 * @retuns -1
 *  EinaPlaylist* l: [self]
 *  gint ret: Next position to active stream
 */
gint eina_playlist_get_next(EinaPlaylist *l);

/*
 * @eina_playlist_go_prev
 * @Changes active stream to the previous available one. If there is previous stream, returns
 * @FALSE, otherwise returns FALSE.
 * @This is equivalent to do:
 * @ i = eina_playlist_get_prev(l);
 * @ eina_playlist_set_current(l, i);
 *  EinaPlaylist* l: [self]
 *  gboolean ret: If EinaPlaylist was able to change to previous stream
 */
gboolean eina_playlist_go_prev (EinaPlaylist *l);

/*
 * @eina_playlist_go_next
 * @Changes active stream to the next available one. If there is next stream, returns
 * @FALSE, otherwise returns FALSE.
 * @This is equivalent to do:
 * @ i = eina_playlist_get_next(l);
 * @ eina_playlist_set_current(l, i);
 *  EinaPlaylist* l: [self]
 *  gboolean ret: If EinaPlaylist was able to change to next stream
 */
gboolean eina_playlist_go_next (EinaPlaylist *l);

/*
 * @eina_playlist_go_nth
 * @Changes active stream to the stream 'pos'.
 * @If pos is equal to -1, active is set to -1
 *  EinaPlaylist* l: [self]
 *  gboolean ret: If EinaPlaylist was able to change to stream 'pos'. If pos is equal to -1, returns TRUE
 */
gboolean eina_playlist_go_nth (EinaPlaylist *l, gint pos);

/*
 * @eina_playlist_swap
 * @Swaps streams at positions 'a' and 'b', returns TRUE on successful operation, FALSE otherwise
 *  EinaPlaylist* l: [self]
 *  guint a: First element to swap
 *  guint b: Second element to swap
 *  gboolean ret: Successful status of operation
 */
gboolean eina_playlist_swap(EinaPlaylist *l, guint a, guint b);

/* 
 * @eina_playlist_randomize
 * @Rebuilds internal random list of playlist
 *  EinaPlaylist* l: [self]
 */
void eina_playlist_randomize(EinaPlaylist *l);

#if 0
/* 
 * @eina_playlist_find
 * @Finds the position in playlist of one stream
 * @Returns -1 if the stream is not found inside playlist
 *  EinaPlaylist* l: [self]
 *  EinaStream* stream: EinaStream to find
 *  gint ret: Position of the stream in playlist or -1 if not found
 */
gint eina_playlist_find(EinaPlaylist *l, EinaStream *stream);
#endif

/*
 * @eina_playlist_print
 * @Prints to stdout the internal playlist
 *  EinaPlaylist* l: [self]
 */
void eina_playlist_print(EinaPlaylist *l);

/*
 * @eina_playlist_print_random
 * @Prints to stdout the internal playlist in random mode
 *  EinaPlaylist* l: [self]
 */
void eina_playlist_print_random(EinaPlaylist *l);

G_END_DECLS

#endif /* __LIBEINA_PLAYLIST_H__ */
