#include <glib/gprintf.h>
#include <libeina/core.h>
#include <string.h>
#include <libeina/i18n.h>

/* Define or domain and GQuark */
#define EINA_CORE_DOMAIN "EinaCore"
static GQuark EinaCoreDomainQuark;

/* :r ! grep EINA_CORE_ERROR elib-core.h
	EINA_CORE_ERROR_NO_PIPELINE,
	EINA_CORE_ERROR_NO_SRC,
	EINA_CORE_ERROR_NO_SPIDER,
	EINA_CORE_ERROR_NO_SINK,
	EINA_CORE_ERROR_NO_VOLUME,
	EINA_CORE_ERROR_NO_VOLUME_DPARAM,
	EINA_CORE_ERROR_NO_MUTE_DPARAM
 */
/*
static gchar *EinaCoreErrors[] = {
	"cannot create pipeline thread",
	"src '%s' is not available",
	"spider element is not available",
	"sink '%s' is not available",
	"volume element is not available",
	"cannot add dparam volume",
	"cannot add dparam mute",
	NULL
};
*/

struct _EinaCore {
	/* EinaCore is not owner of EinaStream */
	const EinaStream *stream;

	GstElement       *pipeline;
	GstElement       *source;
	GstElement       *spider;
	GstElement       *volume;
	GstDParamManager *vol_dpman,  *mute_dpman;
	GstDParam        *vol_dparam, *mute_dparam;
	GstElement       *sink;
};

/*
 * Initialize some elements at Gstreamer(lowest) level
 */
void eina_core_init(gint *argc, gchar **argv[]) {
	gst_init(argc, argv);
	gst_control_init(argc, argv);
	EinaCoreDomainQuark = g_quark_from_static_string(EINA_CORE_DOMAIN);
}

/* Create EinaCore object, ready to be used */
EinaCore * eina_core_new(GError **err) {
	EinaCore *self;
	gchar *input, *output;
	gint cachesize, volume;

	input     = "fakesrc";
	output    = "fakesink";
	cachesize = 8192;
	volume    = 50;
	
	/* Build pipeline
	 * 
	 * pipeline looks something like this:
	 * (file|http|gnomevfs)src ! spider ! volume ! (oss|esd)sink
	 */
	self = g_new0(EinaCore, 1);
	
	self->pipeline   = gst_thread_new("pipeline");
	if ( self->pipeline == NULL ) {
		if ( *err != NULL )
			*err = g_error_new(EinaCoreDomainQuark,
				EINA_CORE_ERROR_NO_PIPELINE,
				_("cannot create pipeline thread"));
		return NULL;
	}

	self->source     = gst_element_factory_make (input, "src");
	if ( self->source == NULL ) {
		if ( *err != NULL )
			*err = g_error_new(EinaCoreDomainQuark,
				EINA_CORE_ERROR_NO_SRC,
				_("src '%s' is not available"), input);
		return NULL;
	}
	eina_core_set_stream(self, NULL);

	/*
	self->cache      = gst_element_factory_make ("cache", "cache");
	g_return_val_if_fail (self->cache    != NULL, NULL);
	*/

	self->spider     = gst_element_factory_make ("spider", "decoder");
	if ( self->spider == NULL ) {
		if ( *err != NULL )
			*err = g_error_new(EinaCoreDomainQuark,
				EINA_CORE_ERROR_NO_SPIDER,
				_("spider element is not available"));
		return NULL;
	}

	self->sink       = gst_element_factory_make (output, "sink");
	if ( self->sink == NULL ) {
		if ( *err != NULL )
			*err = g_error_new(EinaCoreDomainQuark,
				EINA_CORE_ERROR_NO_SINK,
				_("sink '%s' is not available"), output);
		return NULL;
	}

	self->volume = gst_element_factory_make("volume","volume");
	if ( self->sink == NULL ) {
		if ( *err != NULL )
			*err = g_error_new(EinaCoreDomainQuark,
				EINA_CORE_ERROR_NO_VOLUME,
				_("volume element is not available"));
		return NULL;
	}

	/* Volume DParam stuff */
	self->vol_dpman = gst_dpman_get_manager(self->volume);
	gst_dpman_set_mode(self->vol_dpman, "synchronous");
	self->vol_dparam = gst_dparam_new(G_TYPE_DOUBLE);
	if (!gst_dpman_attach_dparam(self->vol_dpman, "volume", self->vol_dparam)) {
		if ( *err != NULL )
			*err = g_error_new(EinaCoreDomainQuark,
				EINA_CORE_ERROR_NO_VOLUME_DPARAM,
				_( "cannot add dparam volume"));
		return NULL;
	}
	eina_core_set_volume(self, volume);

	/* Mute DParam stuff */
	self->mute_dpman = gst_dpman_get_manager(self->volume);
	gst_dpman_set_mode(self->mute_dpman, "synchronous");
	self->mute_dparam = gst_dparam_new(G_TYPE_INT);
	if (!gst_dpman_attach_dparam (self->mute_dpman, "mute", self->mute_dparam)){
		if ( *err != NULL )
			*err = g_error_new(EinaCoreDomainQuark,
				EINA_CORE_ERROR_NO_MUTE_DPARAM,
				_( "cannot add dparam mute"));
		return NULL;
	}
	eina_core_set_mute(self, FALSE);
	
	gst_bin_add_many (GST_BIN(self->pipeline),
		self->source,
		// self->cache,
		self->spider,
		self->volume,
		self->sink,
		NULL);
	gst_element_link_many (self->source,
		// self->cache,
		self->spider,
		self->volume,
		self->sink,
		NULL);
	gst_bin_auto_clock (GST_BIN(self->pipeline));
	return self;
}

/*
 * Deleting the core
 */
void eina_core_unref(EinaCore *p) {
	eina_core_set_state(p, EINA_CORE_STOP);

	/* 
	 * Dont unlink elements, remove implies unlink 
	 * if we try to unlink and remove we can get a segfault
	 */
	gst_bin_remove(GST_BIN(p->pipeline), p->source);
	// gst_bin_remove(GST_BIN(p->pipeline), p->cache);
	gst_bin_remove(GST_BIN(p->pipeline), p->spider);
	gst_bin_remove(GST_BIN(p->pipeline), p->volume);
	gst_bin_remove(GST_BIN(p->pipeline), p->sink);

	gst_object_unref (GST_OBJECT(p->pipeline));
	g_free(p);
}

void eina_core_signal_connect(EinaCore *p, gchar *signal, GCallback function, gpointer data) {
	if ( !strcmp(signal, "eos"))
		g_signal_connect(GST_ELEMENT(p->source), signal, function, data);
	else
		g_signal_connect(GST_ELEMENT(p->pipeline), signal, function, data);
}

gboolean eina_core_set_input(EinaCore *p, gchar* driver, GError **err) {
	GstElement *newinput;
	
	newinput = gst_element_factory_make(driver, "src");
	if ( newinput == NULL ) {
		if ( err != NULL )
			*err = g_error_new(EinaCoreDomainQuark,
				EINA_CORE_ERROR_NO_SRC,
				_("src '%s' is not available"), driver);
		return FALSE;
	}

	/* Rebuild pipeline */
	gst_element_unlink(p->source, p->spider);
	gst_bin_remove(GST_BIN(p->pipeline), p->source);
	
	p->source = newinput;
	
	gst_bin_add(GST_BIN(p->pipeline), p->source);
	gst_element_link(p->source, p->spider);

	return TRUE;
}

gboolean eina_core_set_output(EinaCore *p, gchar *driver, GError **err) {
	GstElement *newoutput;

	newoutput = gst_element_factory_make(driver, "sink");
	if ( newoutput == NULL ) {
		if ( err != NULL )
			*err = g_error_new(EinaCoreDomainQuark,
				EINA_CORE_ERROR_NO_SINK,
				_("sink '%s' is not available"), driver);
		return FALSE;
	}

	/* Rebuild pipeline */
	gst_element_unlink(p->volume, p->sink);
	gst_bin_remove(GST_BIN(p->pipeline), p->sink);
	
	p->sink = newoutput;
	
	gst_bin_add(GST_BIN(p->pipeline), p->sink);
	gst_element_link(p->volume, p->sink);

	return TRUE;
}

/*
 * Get / Set stream of the core
 */ 
EinaCoreStateReturn eina_core_set_stream(EinaCore *p, const EinaStream *stream) {
	EinaCoreState prev_state;

	/* Save state to restore it when stream change was done */
	prev_state = eina_core_get_state(p);
	eina_core_set_state(p, EINA_CORE_STOP);

	if ( stream != NULL ) {
		/* Unlink spider element */
		gst_element_unlink(p->source, p->spider);
		gst_element_unlink(p->spider, p->volume);
		gst_bin_remove(GST_BIN(p->pipeline), p->spider);

		/* Regenerate it */
		p->spider = gst_element_factory_make ("spider", "decoder");
		gst_element_link(p->source, p->spider);
		gst_element_link(p->spider, p->volume);
		gst_bin_add(GST_BIN(p->pipeline), p->spider);
	
		g_object_set (G_OBJECT(p->source), "location", eina_stream_get_location(stream), NULL);
	} else {
		 g_object_set (G_OBJECT(p->source), "location", NULL, NULL);
	}

	/* Restore initial state */
	p->stream = stream;
	return eina_core_set_state(p, prev_state);
}

const EinaStream *eina_core_get_stream(EinaCore *p) {
	return p->stream;
}

/*
 * Get / Set state of core, using an attribute may be more efficient ?
 */
EinaCoreState eina_core_get_state(EinaCore *p) {
	return gst_element_get_state(p->pipeline);
}

EinaCoreStateReturn eina_core_set_state(EinaCore *p, EinaCoreState state) {
	EinaCoreStateReturn r = EINA_CORE_SUCCESS;
	
	/* If we are already in this state return */
	if ( eina_core_get_state(p) == state )
		return EINA_CORE_SUCCESS;

	switch ( state ) {
		case EINA_CORE_VOID:
		case EINA_CORE_READY:
			r = EINA_CORE_SUCCESS;
			break;

		case EINA_CORE_STOP:
		case EINA_CORE_PAUSE:
		case EINA_CORE_PLAY:
			r = gst_element_set_state(p->pipeline, state);
			break;
	}
	
	return r;
}

/*
 * Get / Set volume or mute info
 */
gboolean eina_core_set_volume(EinaCore *p, gint vol) {
	GValue gvol;

	g_return_val_if_fail (p != NULL, FALSE);
	if ( vol > 100 ) vol = 100;
	if ( vol < 0   ) vol = 0;
	
	memset(&gvol, 0, sizeof(gvol));
	g_value_init(&gvol, G_TYPE_DOUBLE);
	g_value_set_double(&gvol, vol / (gdouble) 50);
	g_object_set_property(G_OBJECT(p->vol_dparam), "value_double", &gvol);

	return TRUE;
}

gint eina_core_get_volume (EinaCore *p) {
	GValue gval;

	g_return_val_if_fail (p != NULL, -1);

	memset(&gval, 0, sizeof(gval));
	g_value_init(&gval, G_TYPE_DOUBLE);
	g_object_get_property(G_OBJECT(p->vol_dparam), "value_double", &gval);
	return (gint) (g_value_get_double(&gval) * 50);
}

gboolean eina_core_set_mute(EinaCore *p, gboolean mute) {
	GValue gvol;

	g_return_val_if_fail (p != NULL, FALSE);

	memset(&gvol, 0, sizeof(gvol));
	g_value_init(&gvol, G_TYPE_INT);
	g_value_set_int(&gvol, mute);
	g_object_set_property(G_OBJECT(p->mute_dparam), "value_int", &gvol);

	return TRUE;
}

gboolean eina_core_get_mute (EinaCore *p) {
	GValue gval;

	g_return_val_if_fail (p != NULL, FALSE);

	memset(&gval, 0, sizeof(gval));
	g_value_init(&gval, G_TYPE_INT);
	g_object_get_property(G_OBJECT(p->mute_dparam), "value_int", &gval);
	if ( g_value_get_int(&gval) )
		return TRUE;
	else
		return FALSE;
}

/*
 * Get info about length, position and seek over the stream
 * Idea: Insert a second parameter for set format? Think on large streams,
 *       gint64 overflows
 */ 
gint64 eina_core_get_position(EinaCore *p) {
	GstFormat format;
	EinaCoreState state;
	gint64 pos;

	if ( p->stream == NULL ) return 0;

	format = GST_FORMAT_TIME;
	state = eina_core_get_state(p);
	if ( (state != EINA_CORE_PLAY) && (state != EINA_CORE_PAUSE) )
		pos = 0;
	else if ( !gst_element_query(p->sink, GST_QUERY_POSITION, &format, &pos) ) {
		pos = -1;
	}
	return pos;
}

gboolean eina_core_set_position(EinaCore *p, gint64 pos) {
	GstEvent *ev;

	if ( p->stream == NULL ) return FALSE;

	ev = gst_event_new_seek (GST_FORMAT_TIME |
		GST_SEEK_METHOD_SET |
		GST_SEEK_FLAG_FLUSH, pos);

	return gst_element_send_event(p->sink, ev);
}

gint64 eina_core_length(EinaCore *p) {
	GstFormat format = GST_FORMAT_TIME;
	gint64 duration = 0;

	if ( p->stream == NULL ) return 0;

	gst_element_query (p->sink, GST_QUERY_TOTAL, &format, &duration);
	return duration;
}

