#include <glib.h>
#include <glib/gprintf.h>
#include <gst/gst.h>
#include <gst/control/control.h>
#include <libeina/metadata.h>
#include <libeina/tag.h>

struct _EinaMetadataEngine {
	gboolean eos, handoff, error;
	GstElement *pipeline, *source, *typefind, *sink, *spider;
	EinaStream *stream;
};

static void _eina_metadata_engine_load_tag (const GstTagList *list, const gchar *tag, EinaMetadataEngine *self) {
	const GValue *tag_val;
	gchar *tag_val_string;
	GType tag_type;
	EinaTag eina_tag;

	eina_tag       = eina_tag_get_tag_from_string(tag);
	tag_type       = eina_tag_get_type(eina_tag);
	tag_val        = gst_tag_list_get_value_index (list, tag, 0);
	tag_val_string = NULL;
	
	if ( eina_tag == EINA_TAG_INVALID ) {
		g_warning("%s: invalid tag '%s'\n", __FUNCTION__, tag);
		return;
	}
	
	switch ( tag_type ) {
		case G_TYPE_STRING:
			tag_val_string = g_value_dup_string(tag_val);
			break;
		
		case G_TYPE_INT:
			tag_val_string = g_strdup_printf("%d", (g_value_get_int(tag_val)));
			break;
	
		case G_TYPE_UINT:
			tag_val_string = g_strdup_printf("%u", (g_value_get_uint(tag_val)));
			break;
			
		case G_TYPE_UINT64:
			tag_val_string = g_strdup_printf("%llu", (long long unsigned)(g_value_get_uint64(tag_val)));
			break;

		default:
			g_warning("%s: Tag '%s' has non handle type '%s' please report this as a bug\n", __FUNCTION__, tag, G_VALUE_TYPE_NAME(tag_val));
			break;
	}

	if ( tag_val_string != NULL ) {
		eina_stream_set_metadata(self->stream, eina_tag, tag_val_string);
		g_free(tag_val_string);
	}
}

static void _eina_metadata_engine_tag_found_cb(GObject *pipeline, GstElement *source, GstTagList *tags, EinaMetadataEngine *self) {
	gst_tag_list_foreach (tags, (GstTagForeachFunc) _eina_metadata_engine_load_tag, self);
}

static void _eina_metadata_engine_error_cb(GstElement *element, GstElement *source, GError *error, gchar *debug, EinaMetadataEngine *self) {
	self->error = TRUE;
	g_printf("Error: %s\n", error->message);
}

static void _eina_metadata_engine_typefind_cb (GstElement *typefind, guint probability, GstCaps *caps, EinaMetadataEngine *self) {
	return;

	if (gst_caps_get_size (caps) > 0)
		g_printf("found type: %s\n", gst_structure_get_name (gst_caps_get_structure (caps, 0)));
}

static void _eina_metadata_engine_handoff_cb(GstElement *fakesink, GstBuffer *buf, GstPad *pad, EinaMetadataEngine *self) {
	self->handoff = TRUE;
	return;
}

static void _eina_metadata_engine_eos_cb(GstElement *element, EinaMetadataEngine *self) {
	if ( self->eos ) {
		return;
	}
	gst_element_set_state(self->pipeline, GST_STATE_NULL);
	self->eos = TRUE;
}

EinaMetadataEngine *eina_metadata_engine_new(gchar *src) {
	EinaMetadataEngine *self;
	GstCaps *filtercaps = NULL;

	self = g_new0(EinaMetadataEngine, 1);
	self->error = self->eos = self->handoff = FALSE;

	/* Setup pipeline */
	self->pipeline = gst_pipeline_new("pipeline");
	g_signal_connect(
			G_OBJECT(self->pipeline), "found-tag",
			G_CALLBACK(_eina_metadata_engine_tag_found_cb), self);
	g_signal_connect(
			G_OBJECT(self->pipeline), "error",
			G_CALLBACK(_eina_metadata_engine_error_cb), self);

	/* Setup source */
	self->source   = gst_element_factory_make(src, "src");
	g_object_set(G_OBJECT(self->source), "location", NULL, NULL);
	
	/* Setup typefind */
	self->typefind = gst_element_factory_make("typefind", "typefind");
	g_signal_connect(G_OBJECT(self->typefind), "have_type", G_CALLBACK(_eina_metadata_engine_typefind_cb), self);
	
	/* Setup decoder */
	self->spider   = gst_element_factory_make("spider",  "decoder");
	
	/* Setup sink */
	self->sink     = gst_element_factory_make("fakesink", "sink");
	g_object_set    (G_OBJECT(self->sink), "signal-handoffs", TRUE, NULL);
	g_signal_connect(G_OBJECT(self->sink), "handoff", G_CALLBACK(_eina_metadata_engine_handoff_cb), self);
	g_signal_connect(G_OBJECT(self->sink), "eos",     G_CALLBACK(_eina_metadata_engine_eos_cb),     self);

	/* Put all together and play */
	gst_bin_add_many(GST_BIN(self->pipeline), self->source, self->typefind, self->spider, self->sink, NULL);
	gst_element_link_many(self->source, self->typefind, self->spider, NULL);

	filtercaps = gst_caps_new_simple ("audio/x-raw-int", NULL);
	gst_element_link_filtered (self->spider, self->sink, filtercaps);
	gst_caps_free (filtercaps);
	
	gst_bin_auto_clock(GST_BIN(self->pipeline));

	return self;
}

void eina_metadata_engine_unref(EinaMetadataEngine *self) {
	gst_element_set_state(self->pipeline, GST_STATE_NULL);
	gst_element_unlink_many(self->source, self->typefind, self->spider, self->sink, NULL);
	gst_bin_remove_many(GST_BIN(self->pipeline), self->source, self->typefind, self->spider, self->sink, NULL);
	
	g_free(self);
}

gboolean eina_metadata_engine_parse(EinaMetadataEngine *self, EinaStream *stream) {
	GstCaps *filtercaps;
	self->stream = stream;
	
	g_object_set(G_OBJECT(self->source), "location", eina_stream_get_location(self->stream), NULL);


	self->handoff = self->eos = self->error = FALSE;
	self->error   = !gst_element_set_state(self->pipeline, GST_STATE_PLAYING);
	while (gst_bin_iterate(GST_BIN(self->pipeline)) && !self->eos && !self->handoff && !self->error) {
		;
	}
	
	/* Clean pipeline */
	gst_element_set_state(self->pipeline, GST_STATE_NULL);
	g_object_set(G_OBJECT(self->source), "location", NULL, NULL);
	
	/* Delete any reference of spider inside pipeline */
	gst_element_unlink(self->typefind, self->spider);
	gst_element_unlink(self->spider, self->sink);
	gst_bin_remove(GST_BIN(self->pipeline), self->spider);
	
	/* Regenerate spider element */
	self->spider = gst_element_factory_make ("spider", "decoder");

	/* Insert again into pipeline */
	gst_bin_add(GST_BIN(self->pipeline), self->spider);
	gst_element_link(self->typefind, self->spider);
	filtercaps = gst_caps_new_simple ("audio/x-raw-int", NULL);
	gst_element_link_filtered (self->spider, self->sink, filtercaps);
	gst_caps_free (filtercaps);

	return !self->error;
}

