#include <glib/gprintf.h>
#include <glib-object.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <libeina/player.h>
#include <libeina/player-priv.h>
#include <libeina/metadata.h>
#include <libeina/i18n.h>

#define EINA_PLAYER_DOMAIN "EinaPlayer"
static GQuark EinaPlayerDomainQuark;

/*
 * Helper functions for add streams 
 *
 * add_dir and add_m3u are wrappers to add_file
 */
gboolean eina_player_add_file(EinaPlayer *p, EinaStream *stream);
gboolean eina_player_add_dir (EinaPlayer *p, EinaStream *stream);
gboolean eina_player_add_m3u (EinaPlayer *p, EinaStream *stream);

/*
 * Init player system
 */
void eina_player_init(int *argc, char **argv[]) {
	g_type_init();
	eina_core_init(argc, argv);
	EinaPlayerDomainQuark =  g_quark_from_static_string(EINA_PLAYER_DOMAIN);
}

/*
 * Create new player
 */
EinaPlayer *eina_player_new (GError **err) {
	EinaPlayer *p;

	p = g_new0(EinaPlayer, 1);
	
	p->core     = eina_core_new(err);
	p->core_aux = NULL;
	g_return_val_if_fail ( *err == NULL, NULL);

	p->pl     = eina_playlist_new();
	p->metadata_engine = eina_metadata_engine_new("fakesrc");
	p->eos    = FALSE;
	p->slots  = NULL;
	p->state  = EINA_STATE_STOP;
	p->mimedetector = (magic_t) magic_open(
						MAGIC_MIME |
						MAGIC_PRESERVE_ATIME |
						MAGIC_SYMLINK);
	magic_load(p->mimedetector, "/usr/share/misc/file/magic");

	eina_core_signal_connect(p->core,
		"eos",
		G_CALLBACK(eina_player_trap_eos),
		(gpointer) p);

	p->watcher_id = g_timeout_add(100, eina_player_eos_watcher,    p);

	return p;
}

/* Finalize player*/
void eina_player_unref(EinaPlayer *p) {
	GList *pl;
	
	magic_close(p->mimedetector);
	g_source_remove(p->watcher_id);
	eina_core_unref(p->core);

	/* Delete each stream in playlist */
	pl = eina_playlist_get_playlist(p->pl);
	while ( pl ) {
		eina_stream_unref((EinaStream *) pl->data);
		pl = pl->next;
	}
	eina_playlist_unref(p->pl);

	eina_metadata_engine_unref(p->metadata_engine);
	g_list_free(p->slots);
	g_free(p);
}

gboolean eina_player_set_input(EinaPlayer *p, gchar *input, GError **err) {
	if ( eina_core_get_state(p->core) != EINA_CORE_STOP ) {
		if ( err != NULL )
			*err = g_error_new(EinaPlayerDomainQuark,
						EINA_PLAYER_ERROR_SET_INPUT,
						_("cannot change input if state != stop,  implement delaying\n"));
		return FALSE;
	}

	if ( !eina_core_set_input(p->core, input, err) )
		return FALSE;

	/* Copy input */
	if ( p->src != NULL ) g_free(p->src);
	p->src = g_strdup(input);

	/* Rebuild metadata engine */
	eina_metadata_engine_unref(p->metadata_engine);
	p->metadata_engine = eina_metadata_engine_new(input);

	eina_core_signal_connect(p->core,
		"eos",
		G_CALLBACK(eina_player_trap_eos),
		(gpointer) p);

	return TRUE;
}

const gchar *eina_player_get_input(EinaPlayer *p) {
	return p->src;
}

gboolean eina_player_set_output(EinaPlayer *p, gchar *output, GError **err) {
	if ( eina_core_get_state(p->core) != EINA_CORE_STOP ) {
		if ( err != NULL )
			*err = g_error_new(EinaPlayerDomainQuark,
						EINA_PLAYER_ERROR_SET_OUTPUT,
						_("cannot change output if state != stop,  implement delaying\n"));
		return FALSE;
	}

	if ( !eina_core_set_output(p->core, output, err) )
		return FALSE;

	/* Copy input */
	if ( p->sink != NULL ) g_free(p->sink);
	p->sink = g_strdup(output);

	return TRUE;
}

const gchar *eina_player_get_output(EinaPlayer *p) {
	return p->sink;
}

/*
 * Slot funcions
 *
 * Add/remove slots from player
 */
guint eina_player_slot_add(EinaPlayer *p, EinaSlot slot) {
	EinaSlot *np;
	np = g_new0(EinaSlot, 1);
	memcpy(np, &slot, sizeof(EinaSlot));
	
	p->slots = g_list_append(p->slots, np);
	
	return g_list_index(p->slots, (gconstpointer) np);
}

void eina_player_slot_remove(EinaPlayer *p, gint slot_id) {
	gpointer slot;
	slot = g_list_nth_data(p->slots, slot_id);
	g_return_if_fail(slot != NULL);
					
	p->slots = g_list_remove(p->slots, slot);
	g_free(slot);
}

/*
 *
 * Add functions, some wrappers and the real add
 */

/* Adds an uri to player */
gboolean eina_player_add_uri(EinaPlayer *p, gchar *uri) {
	EinaStream *stream;
	stream = eina_stream_new(uri);
	g_return_val_if_fail(stream != NULL, FALSE);
	return eina_player_add(p, stream);
}

gboolean eina_player_add(EinaPlayer *p, EinaStream *stream) {
	gboolean add_ok = FALSE;
	gboolean is_audio, is_ogg;
	const gchar *mimetype;
	const gchar *filename;
	
	if ( !eina_stream_is_local(stream) ) {
		/* Something on GnomeVFS/KIOSlave/... add */
		return eina_player_add_file(p, stream);
	} else {
		filename = eina_stream_get_location(stream);
		mimetype = eina_priv_get_mime(p, filename); /* Stream is filesystem encoding */
		if ( mimetype == NULL )
			return FALSE;

		is_audio = g_str_has_prefix(mimetype, "audio/");
		is_ogg   = !g_ascii_strcasecmp("application/ogg", mimetype);

		/* Filter by category */
		if ( g_file_test(filename, G_FILE_TEST_IS_DIR) ) {
			add_ok |= eina_player_add_dir(p, stream);
		}

		else if ( is_audio || is_ogg ) {
			add_ok |= eina_player_add_file(p, stream);
		}

		else if ( !g_ascii_strcasecmp("text/plain", mimetype) ) {
			add_ok |= eina_player_add_m3u(p, stream);
		} 

		return add_ok;
	}
}

gboolean eina_player_add_dir(EinaPlayer *p, EinaStream *stream) {
	EinaStream *n_stream;
	GDir *directory;
	GSList *dirs = NULL, *files = NULL;
	gpointer elem;
	const gchar *entry, *dirname;
	gchar *fullname;
	gint i;
	gboolean add_ok = FALSE;

	dirname = eina_stream_get_location(stream);
	
	if ( ! g_file_test(dirname, G_FILE_TEST_IS_DIR) ) 
		return FALSE;
	
	/* Read entries, distinguishing files and dirs */
	directory = g_dir_open(dirname, 0, NULL);
	while ( (entry = g_dir_read_name(directory)) ) {
		fullname = g_build_filename(dirname, entry);
		if ( g_file_test(fullname, G_FILE_TEST_IS_DIR) )
			dirs  = g_slist_append(dirs,  fullname);
		else
			files = g_slist_append(files, fullname);
	}
	g_dir_close(directory);
	
	/* Order the lists */
	dirs  = g_slist_sort(dirs,  (GCompareFunc) g_ascii_strcasecmp);
	files = g_slist_sort(files, (GCompareFunc) g_ascii_strcasecmp);
	
	/* Add both lists */
	for ( i = 0; (elem = g_slist_nth_data(dirs,i)); i++) {
		n_stream = eina_stream_new(elem);
		add_ok |= eina_player_add_dir(p, n_stream);
		eina_stream_unref(n_stream);
	}
	for ( i = 0; (elem = g_slist_nth_data(files,i)); i++) {
		n_stream = eina_stream_new(elem);
		add_ok |= eina_player_add(p, n_stream);
	}

	g_slist_free(files);
	g_slist_free(dirs);

	return add_ok;
}

gboolean eina_player_add_file(EinaPlayer *p, EinaStream *stream) {
	gboolean add_ok;
	gint total;

	if ( !eina_metadata_engine_parse(p->metadata_engine, stream) ) {
		/* XXX Posible memleak: who should free this useless stream? */
		g_warning("Error parsing metadata for %s\n", eina_stream_get_location(stream));
		return FALSE; /* Unsupported type or unreadable stream */
	}
	
	add_ok = eina_playlist_add(p->pl, stream);
	total = eina_playlist_total(p->pl) - 1;

	_eina_player_emit_signal(p, EINA_SIGNAL_ADD, stream, (gpointer) total);
	return TRUE;
}

gboolean eina_player_add_m3u(EinaPlayer *p, EinaStream *stream) {
	EinaStream *n_stream;
	gboolean add_ok = FALSE;
	const gchar *filename;
	gchar *contents = NULL;
	gchar **lines = NULL;
	gchar *fullpath = NULL, *dirname = NULL;
	gint i;
	
	filename = eina_stream_get_location(stream);
	dirname = g_path_get_dirname(filename);
	if ( !g_file_get_contents(filename, &contents, NULL, NULL) )
		return FALSE;
	
	lines = g_strsplit(contents, "\n", 0);
	for ( i = 0; lines[i] != NULL; i++) {
		if ( lines[i][strlen(lines[i])] == '\r' )
			lines[i][strlen(lines[i])] = '\0';

		if (
			g_ascii_isalnum(lines[i][0]) ||
			lines[i][0] == '.'           ||
			lines[i][0] == '/'
			) {
			if ( lines[i][0] == '/' ) {
				fullpath = g_strdup(lines[i]);
			} else {
				fullpath = g_build_filename(dirname, lines[i], NULL);
			}
			n_stream = eina_stream_new(fullpath);
			add_ok |= eina_player_add(p, n_stream);
			g_free(fullpath);
		}
	}
	
	g_strfreev(lines);
	g_free(dirname);
	g_free(contents);

	return add_ok;
}

/* Remove song at position 'pos' */
void eina_player_del(EinaPlayer *p, gint pos) {
	EinaStream *deleted_stream;
	gint total;
	gboolean is_actual;
	EinaCoreState state = EINA_CORE_FAILURE;

	if ( (pos < 0) ||  pos > eina_player_total(p) )
		return;

	is_actual = (pos == eina_playlist_get_current(p->pl));

	/* Deleting active element requires extra effort */
	if ( is_actual ) {
		state = eina_core_get_state(p->core);
		/* XXX: Try to delete this 'stop' */
		eina_player_stop(p);
		eina_player_next(p);
	}

	deleted_stream = eina_stream_clone((EinaStream *) eina_player_nth_stream(p, pos));
	eina_playlist_del(p->pl, pos);
	_eina_player_emit_signal(p, EINA_SIGNAL_DEL, deleted_stream, (gpointer) pos);
	eina_stream_unref(deleted_stream);
	
	total = eina_playlist_total(p->pl);

	if ( is_actual )
		eina_core_set_state(p->core, state);
}

/* Clear & reset player */
void eina_player_clear(EinaPlayer *p) {
	GList *pl;

	eina_player_stop(p);
	pl = eina_playlist_get_playlist(p->pl);
	while ( pl ) {
		eina_stream_unref((EinaStream *) pl->data);
		pl = pl->next;
	}
	eina_playlist_clear(p->pl);
	eina_core_set_stream(p->core, NULL);
	eina_core_set_state(p->core, EINA_CORE_STOP);
	p->state = EINA_STATE_STOP;
	_eina_player_emit_signal(p, EINA_SIGNAL_CLEAR, NULL, NULL);
}

/*
 * Start player
 * XXX  Rewrite code or wrap eina_player_nth ?
 */
gboolean eina_player_play (EinaPlayer *p) {
	const EinaStream *stream;
	EinaCoreStateReturn ret;
	gint                current, total;

	if ( eina_playlist_total(p->pl) <= 0 )
		return FALSE;

	if ( eina_core_get_state(p->core) == EINA_CORE_PLAY )
		return TRUE;
	
	if ( eina_core_get_state(p->core) == EINA_CORE_PAUSE ) {
		eina_core_set_state(p->core, EINA_CORE_STOP);
	}

	/* Data about present playlist */
	current = eina_playlist_get_current(p->pl);
	total  = eina_player_total(p);
	stream = eina_player_nth_stream(p, current);

	eina_player_nth(p, current);
	ret = eina_core_set_state(p->core, EINA_CORE_PLAY);
	
	if ( ret == EINA_CORE_FAILURE ) {
		g_warning("Cannot set state to play\n");
		eina_core_set_state(p->core, EINA_CORE_STOP);
		return FALSE;
	}
	else {
		p->state = EINA_STATE_PLAY;
		_eina_player_emit_signal(p, EINA_SIGNAL_PLAY, (gpointer) stream, (gpointer) current);
		return TRUE;
	}
}

/* Pause player */
gboolean eina_player_pause (EinaPlayer *p) {
	gboolean lock;
	switch ( eina_core_get_state(p->core)  ) {
		case EINA_CORE_PLAY:
			lock = TRUE;
			p->state = EINA_STATE_PAUSE;
			eina_core_set_state(p->core, EINA_CORE_PAUSE);
			break;

		case EINA_CORE_PAUSE:
			lock = FALSE;
			p->state = EINA_STATE_PLAY;
			eina_core_set_state(p->core, EINA_CORE_PLAY);
			break;

		default:
			return FALSE;
	}

	_eina_player_emit_signal(p, EINA_SIGNAL_PAUSE, (gpointer) lock, NULL);
	return TRUE;
}

/* Stop player */
gboolean eina_player_stop (EinaPlayer *p) {
	p->state = EINA_STATE_STOP;
	eina_core_set_state(p->core, EINA_CORE_STOP);
	_eina_player_emit_signal(p, EINA_SIGNAL_STOP, NULL, NULL);
	return TRUE;
}

/* Move player to previous song */
gboolean eina_player_prev (EinaPlayer *p) {
	gint old, new;
	old = eina_playlist_get_current(p->pl);
	new = eina_playlist_get_prev(p->pl);
	return eina_player_nth(p, new);
}

/* Move player to next song */
gboolean eina_player_next (EinaPlayer *p) {
	gint old, new;
	old = eina_playlist_get_current(p->pl);
	new = eina_playlist_get_next(p->pl);
	return eina_player_nth(p, new);
}

/* Move player to position 'pos' */
gboolean eina_player_nth(EinaPlayer *p, gint pos) {
	gint old;
	const EinaStream *stream;
	EinaState state;

	if ((pos < 0) || (pos >= eina_player_total(p)))
		return FALSE;

	old    = eina_playlist_get_current(p->pl);
	stream = eina_playlist_get_nth(p->pl, pos);

	/* Get state of player */
	state = eina_player_get_state(p);
	
	/* Change values */
	eina_core_set_stream(p->core, stream); 
	eina_playlist_go_nth(p->pl, pos);

	_eina_player_emit_signal(p, EINA_SIGNAL_CHANGE, (gpointer) old, (gpointer) pos);

	/* If prev state was play, we need to emit play signal */
	if ( state == EINA_STATE_PLAY )
		_eina_player_emit_signal(p, EINA_SIGNAL_PLAY, (gpointer) stream, (gpointer) pos);
	
	return TRUE;
}

/* Swap two elements in playlist */
void eina_player_swap(EinaPlayer *p, gint a, gint b) {
	eina_playlist_swap(p->pl, a, b);
	_eina_player_emit_signal(p, EINA_SIGNAL_SWAP, (gpointer) a, (gpointer) b);
}

const EinaStream *eina_player_get_stream(EinaPlayer *p) {
	return eina_core_get_stream(p->core);
}

/* 
 Devuelve el stream de la posicion pos
 */
const EinaStream *eina_player_nth_stream(EinaPlayer *p, gint pos) {
	return eina_playlist_get_nth(p->pl, pos);
}

gint eina_player_current(EinaPlayer *p) {
	return eina_playlist_get_current(p->pl);
}

gint eina_player_total(EinaPlayer *p) {
	return eina_playlist_total(p->pl);
}

void eina_player_set_random(EinaPlayer *p, gboolean value) {
	eina_playlist_set_random(p->pl, value);

	_eina_player_emit_signal(p, EINA_SIGNAL_RANDOM, (gpointer) value, NULL);
}

void eina_player_set_repeat(EinaPlayer *p, gboolean value) {
	eina_playlist_set_repeat(p->pl, value);

	_eina_player_emit_signal(p, EINA_SIGNAL_REPEAT, (gpointer) value, NULL);
}

gboolean eina_player_get_random(EinaPlayer *p) {
	return eina_playlist_get_random(p->pl);
}

gboolean eina_player_get_repeat(EinaPlayer *p) {
	return eina_playlist_get_repeat(p->pl);
}

gint64 eina_player_get_position(EinaPlayer *p) {
	return eina_core_get_position(p->core);
}

void eina_player_set_position(EinaPlayer *p, gint64 pos) {
	gint64 old_pos;
	gint64 len;
	
	len = eina_player_length(p);
	
	if ( pos < 0 )    pos = 0;
	if ( pos > len )  pos = len;
	
	old_pos = eina_player_get_position(p);
	eina_core_set_position(p->core, pos);

	_eina_player_emit_signal(p, EINA_SIGNAL_SEEK, &old_pos, &pos);
}

gint64 eina_player_length(EinaPlayer *p) {
	return eina_core_length(p->core);
}

gboolean eina_player_set_volume(EinaPlayer *p, gint vol) {
	gboolean ret;
	gint old, new;
	
	old = eina_core_get_volume(p->core);
	new = vol;
	
	ret = eina_core_set_volume(p->core, new);

	_eina_player_emit_signal(p, EINA_SIGNAL_VOLUME, (gpointer) old, (gpointer) new);

	return ret;
}

gint eina_player_get_volume(EinaPlayer *p) {
	return eina_core_get_volume(p->core);
}

gboolean eina_player_set_mute(EinaPlayer *p, gboolean mute) {
	return eina_core_set_mute(p->core, mute);
}

gboolean eina_player_get_mute(EinaPlayer *p) {
	return eina_core_get_mute(p->core);
}

GList * eina_player_get_playlist(EinaPlayer *p) {
	return eina_playlist_get_playlist(p->pl);
}

EinaState eina_player_get_state(EinaPlayer *p) {
	return p->state;
}

void eina_player_print_playlist(EinaPlayer *p) {
	eina_playlist_print(p->pl);
}

void eina_player_print_random_playlist(EinaPlayer *p) {
	eina_playlist_print_random(p->pl);
}


