#include <libeina/playlist.h>
#include <glib/gprintf.h>

struct _EinaPlaylist {
    GList *list;
    GList *random_list;
    gboolean repeat;
    gboolean random;
    gint total;
    gint current;
};

/* Private functions */
gint eina_playlist_normal_to_random (EinaPlaylist *l, gint pos);
gint eina_playlist_random_to_normal (EinaPlaylist *l, gint pos);

/*
 * Return position in the random list
 * of the element at position 'pos' in the normal list 
 */
gint eina_playlist_normal_to_random (EinaPlaylist *l, gint pos) {
	gpointer data;
	data = g_list_nth_data(l->list, pos);
	pos = g_list_index(l->random_list, data);
	return pos;
}

/*
 * Return position in the normal list
 * of the element at position 'pos' in the random list 
  */
gint eina_playlist_random_to_normal (EinaPlaylist *l, gint pos) {
	gpointer data;
	data = g_list_nth_data(l->random_list, pos);
	pos = g_list_index(l->list, data);
	return pos;
}

/* Public functions implementation */

/* Create new EinaPlaylist object */
EinaPlaylist * eina_playlist_new(void) {
	EinaPlaylist *self;

	self = g_new0(EinaPlaylist, 1);
	self->list        = NULL;
	self->random_list = NULL;
	eina_playlist_set_repeat(self, FALSE);
	eina_playlist_set_random(self, FALSE);
	self->total       =  0;
	self->current     = -1;
	return self;
}

/* Destroy EinaPlaylist object */
void eina_playlist_unref(EinaPlaylist * l) {
	/* 
	 * Dont do a deep free in playlist.
	 * «We cannot free() what we didn't malloc()»
	 */
	g_list_free(l->list);
	g_list_free(l->random_list);
	g_free(l);
}

GList* eina_playlist_get_playlist(EinaPlaylist *l) {
	return l->list;
}

/* Return number of elements in playlist */
gint eina_playlist_total (EinaPlaylist * l) {
	return l->total;
}

/* Return position of active element */
gint eina_playlist_get_current(EinaPlaylist * l) {
	return l->current;
}

/* Mark element at position 'pos' as active */
void eina_playlist_set_current(EinaPlaylist * l, gint pos) {
	l->current = pos;
}

/* Get random state (TRUE/FALSE) */
gboolean eina_playlist_get_random(EinaPlaylist *l) {
	return l->random;
}

/* Set random ON/OFF */
void eina_playlist_set_random(EinaPlaylist *l, gboolean val) {
	l->random = val;
	if ( l->random ) eina_playlist_randomize(l);
}

/* Get repeat state (TRUE/FALSE) */
gboolean eina_playlist_get_repeat (EinaPlaylist *l) {
	return l->repeat;
}

/* Set repeat ON/OFF */
void eina_playlist_set_repeat(EinaPlaylist *l, gboolean val) {
	l->repeat = val;
}

/* Return element at position 'pos' */
const EinaStream *eina_playlist_get_nth(EinaPlaylist *l, guint pos) {
	return g_list_nth_data(l->list, pos);
}

/* Add URI 'name' to playlist */
gboolean eina_playlist_add(EinaPlaylist *l, const EinaStream *stream) {
	gint randompos;
	
	l->list = g_list_append(l->list, (gpointer) stream);
	if ( l->current == -1 )
		l->current = 0;
	
	randompos = l->total ? g_random_int_range(0, l->total + 1) : 0;
	l->random_list = g_list_insert( l->random_list, (gpointer) stream, randompos);

	l->total++;

	return TRUE;
}

/* Delete element at position 'pos' from playlist */
void eina_playlist_del(EinaPlaylist * l, guint pos) {
	gpointer data;

	if ( (data = g_list_nth_data(l->list, pos)) != NULL ) {
		l->list        = g_list_remove(l->list,        data);
		l->random_list = g_list_remove(l->random_list, data);
		l->total--;

		/* Decrement active element number if the deleted one was previous to it */
		if ( pos <= l->current ) l->current--;

		/*
		g_printf("[libeina::playlist] %s\n", __FUNCTION__);
		g_printf("[libeina::playlist] print follows\n");
		eina_playlist_print(l);
		g_printf("\n");
		*/
	}
}

/* Clear playlist */
void eina_playlist_clear(EinaPlaylist *l) {
	g_list_free(l->list);
	g_list_free(l->random_list);

	l->list        = NULL;
	l->random_list = NULL;

	l->current = -1;
	l->total = 0;
}

gboolean eina_playlist_swap(EinaPlaylist *l, guint a, guint b) {
	gpointer data_a, data_b;
	gint tmp;
	guint total;

	/* Check if the two elements are inside playlist limits */
	total = eina_playlist_total(l);
	if ( (a < 0) || (a >= total) ||
		 (b < 0) || (b >= total) ||
		 (a == b))
		return FALSE;
	
	/* swap(a,b) if b>a; do this to simplify code */
	if ( b > a ) {
		tmp = a; a = b; b = tmp;
	}
	
	data_a = g_list_nth_data(l->list, a);
	data_b = g_list_nth_data(l->list, b);

	/* Delete both elements */
	l->list = g_list_remove(l->list, data_a);
	l->list = g_list_remove(l->list, data_b);

	/* Insert 'b' where 'a' was */
	l->list = g_list_insert(l->list, data_b, a - 1);
	l->list = g_list_insert(l->list, data_a, b);

	if      ( a == l->current ) { l->current = b; }
	else if ( b == l->current ) { l->current = a; }

	return TRUE;
}

/*
 * Return position NEXT to the active,
 * or -1 if it doesn't exist
 */
gint eina_playlist_get_next(EinaPlaylist *l) {
	gint pos;
	gint total, normalpos, randompos;
	
	total     = l->total;
	normalpos = l->current;
	randompos = eina_playlist_normal_to_random(l, l->current);
	
	/* Empty list */
	if ( total == 0 ) {
		return -1;
	}

	/* At end of list */
	if ( l->random && (randompos == total - 1) ) {
		if ( l->repeat ) { return eina_playlist_random_to_normal(l, 0);  }
		else             { return -1; }
	}
	
	if ( !(l->random) && (normalpos == total - 1)) {
		if ( l->repeat ) { return 0; }
		else             { return -1; }
	}

	/* normal advance */
	if ( l->random ) {
		pos = eina_playlist_normal_to_random(l, l->current);
		pos = eina_playlist_random_to_normal(l, pos + 1);
		return pos;
	}
	else
		return l->current+ 1;
}

/*
 * Return position PREVIOUS to the active,
 * or -1 if it doesn't exist
 */
gint eina_playlist_get_prev(EinaPlaylist *l) {
	gint pos;
	gint total, normalpos, randompos;
	
	total     = l->total;
	normalpos = l->current;
	randompos = eina_playlist_normal_to_random(l, l->current);

	/* Empty list */
	if ( total == 0 )
		return -1;

	/* At beginning of list */
	if ( l->random && (randompos == 0) ) {
		if ( l->repeat ) { return eina_playlist_random_to_normal(l, total - 1);  }
		else             { return -1; }
	}
	
	if ( !(l->random) && (normalpos == 0) ) {
		if ( l->repeat ) { return total - 1; }
		else             { return -1; }
	}

	/* normal backwards movement */
	if ( l->random ) {
		pos = eina_playlist_normal_to_random(l, l->current);
		pos = eina_playlist_random_to_normal(l, pos - 1);
		return pos;
	}
	else
		return l->current - 1;
}

/*
 * Goto previous element in play list
 * Return TRUE/FALSE on success/failure
 */
gboolean eina_playlist_go_prev (EinaPlaylist *l) {
	gint prev = eina_playlist_get_prev(l);
	return eina_playlist_go_nth(l, prev);
}

/* 
 * Goto next element in play list
 * Return TRUE/FALSE on success/failure
 */
gboolean eina_playlist_go_next (EinaPlaylist *l) {
	gint next = eina_playlist_get_next(l);
	return eina_playlist_go_nth(l, next);
}

/*
 * Goto element at position 'pos' in play list
 * Return TRUE/FALSE on success/failure
 */
gboolean eina_playlist_go_nth (EinaPlaylist *l, gint pos) {
	if ( pos < 0 ) {
		return FALSE;
	}
	else {
		eina_playlist_set_current(l, pos);
		return TRUE;
	}
}

/* Re-build random list */
void eina_playlist_randomize(EinaPlaylist *l) {
	gint randompos;
	gint i;
	gint total;
	
	g_list_free(l->random_list);
	l->random_list = NULL;

	total = eina_playlist_total(l);
	
	for ( i = 0; i < total; i++ ) {
		if ( g_list_length(l->random_list) )
			randompos = g_random_int_range(0, g_list_length(l->random_list));
		else
			randompos = 0;

		/* Insert at pos randompos-1 to avoid first song always be the 
		 * first in random_list
		*/
		l->random_list = g_list_insert(
							l->random_list,
							g_list_nth_data(l->list, i),
							randompos-1);
	}
}

void eina_playlist_print(EinaPlaylist * l) {
	GList *list;

	// g_printf("[libeina::playlist] %s %p\n", __FUNCTION__, l);
	list = l->list;
	while ( list ) {
		eina_stream_print((EinaStream *) list->data);
		list = list->next;
	}
}

void eina_playlist_print_random(EinaPlaylist * l) {
	GList *list;

	// g_printf("[libeina::playlist] %s %p\n", __FUNCTION__, l);
	list = l->list;
	while ( list ) {
		eina_stream_print((EinaStream *) list->data);
		list = list->next;
	}
}

