#include <string.h>
#include <glib/gprintf.h>
#include <libeina/stream.h>
#define EINA_STREAM_SERIALIZE_CHAR "\1"

struct _EinaStream {
	gchar *stream;   /* URI without protocol */
	
	/* All data in utf8 */
	gchar *uri;        /* Complete URI for stream http:// ftp:// file:// */
	gchar *protocol;   /* The protocol from URI */
	gboolean is_local; /* Is stream is in local disk */
	gchar *basename;   /* Basename of stream a.k.a. filename */

	/* Metadata for stream */
	gchar *metadata[EINA_TAGS];
};

struct TagFmt {
	gchar key;
	EinaTag tag;
};

static struct TagFmt tag_fmt_table [] = {
	{ 'a', EINA_TAG_ARTIST       },
	{ 'b', EINA_TAG_ALBUM        },
	{ 't', EINA_TAG_TITLE        },
	{ 'g', EINA_TAG_GENRE        },
	{ 'n', EINA_TAG_TRACK_NUMBER },
	{ 'd', EINA_TAG_DATE         },
	{   0, EINA_TAG_INVALID      }
};

/* Functions to decode %NN escaped strings */
gchar *url_decode(gchar *str);
gint __hex2int(gchar hex);

EinaStream *eina_stream_new(const gchar *uri) {
	EinaStream *self = g_new0(EinaStream, 1);
	GError *err = NULL;
	gchar *abspath;
	gchar *utf8uri;
	gchar *pwd = NULL;
	gchar *copy_buffer;
	gchar *pointer;

	/* Check Local or remote? */
	if ( strstr(uri, "://") == NULL ) {
		if ( !g_path_is_absolute(uri) ) {
			pwd = g_get_current_dir();
			abspath = g_build_filename(pwd, uri, NULL);
		} else
			abspath = g_strdup(uri);
		
		if   (!g_utf8_validate(uri, -1, NULL)) utf8uri = g_filename_to_utf8(abspath, -1, NULL, NULL, &err);
		else                                   utf8uri = g_strdup(abspath);

		if ( utf8uri == NULL ) {
			g_printf("***: %s\n", err->message);
			return NULL;
		}
		
		//utf8uri = g_strdup(abspath);
	
		self->stream   = g_strdup(uri);
		self->uri      = g_strconcat("file://", utf8uri, NULL);
		self->is_local = TRUE;

		g_free(pwd);
		g_free(abspath);
		g_free(utf8uri);
	} else {
		// g_printf("stream is encoded as uri\n");
		/* Uris always are Utf-8 encoded? */
		self->stream   = g_strdup(uri);
		self->uri      = g_strdup(uri);
		self->is_local = FALSE;
	}

	self->basename = g_path_get_basename(self->uri);
	
	/* Get protocol from uri */
	copy_buffer = g_strdup(self->uri);
	pointer = strstr(copy_buffer, "://");
	pointer[0] = '\0';
	self->protocol = g_strdup(copy_buffer);
	g_free(copy_buffer);

	/* Zero metadata */
	memset(self->metadata, 0, EINA_TAGS * sizeof(gchar *));	

	return self;
}

EinaStream *eina_stream_new_from_string(const gchar *str) {
	return eina_stream_unserialize(str);
}

EinaStream *eina_stream_clone(EinaStream *self) {
	EinaStream *clon;
	EinaTag tag;

	clon = g_new0(EinaStream, 1);
	clon->stream   = g_strdup(self->stream);
	clon->uri      = g_strdup(self->uri);
	clon->protocol = g_strdup(self->protocol);
	clon->is_local = self->is_local;
	clon->basename = g_strdup(self->basename);
	memset(clon->metadata, 0, EINA_TAGS * sizeof(gchar *));	
	
	for ( tag = 0; tag < EINA_TAGS; tag++ ) {
		if ( self->metadata[tag] != NULL )
			clon->metadata[tag] = g_strdup(self->metadata[tag]);
	}

	return clon;
}

void eina_stream_unref(EinaStream *self) {
	EinaTag tag;

	g_free(self->uri);
	g_free(self->stream);
	g_free(self->basename);
	g_free(self->protocol);

	for ( tag = 0; tag < EINA_TAGS; tag++ ) {
		if ( self->metadata[tag] != NULL ) {
			g_free(self->metadata[tag]);
		}
	}
			
	g_free(self);
}

const gchar *eina_stream_get_uri(const EinaStream *self) {
	return self->uri;
}

const gchar *eina_stream_get_location(const EinaStream *self) {
	return self->stream;
}

gboolean eina_stream_is_local(const EinaStream *self) {
	return self->is_local;
}

void eina_stream_set_metadata(EinaStream *self, EinaTag tag, gchar *val) {
	self->metadata[tag] = g_strdup(val);
	// g_printf("%s: %s %s\n", self->uri, eina_tag_get_string_from_tag(tag), self->metadata[tag]);
}

const gchar *eina_stream_get_metadata(const EinaStream *self, EinaTag tag) {
	return self->metadata[tag];
}

gboolean eina_stream_format(const EinaStream *self, const gchar *fmt, gint max_fails, gint flags, gchar **dest) {
	/* 
	 * g_string_append_safe wraps g_string_append allowing to append
	 * a null string without warnings
	 */
#ifdef g_string_append_safe
#undef g_string_append_safe
#endif
#define g_string_append_safe(gs, s) \
	do { \
		if ( s != NULL ) gs = g_string_append(gs, s); \
	} while (0)
	GString *string;
	gint i, j, len;
	gboolean key_found = FALSE;
	gchar *ret = NULL;
	gchar *tag;
	
	/* Temporal string for transformations based on flags */
	gchar *tmp;
	
	len = strlen(fmt);
	string = g_string_new_len(NULL, len * 10);
	
	for ( i = 0; i < len; i++ ) {
		/*
		 * Append characters block
		 *
		 * Current caracters are simple
		 */
		// g_printf("inspecting char '%c': ", fmt[i]);

		/* Current char is not %, add and go next */
		if ( fmt[i] != '%' ) {
			// g_printf("normal\n");
			string = g_string_append_c(string, fmt[i]);
			continue;
		}

		/* We got an '%' but we are at end of string */
		if ( fmt[i+1] == '\0' ) {
			// g_printf("normal\n");
			string = g_string_append_c(string, fmt[i]);
			continue;
		}
		
		/* 
		 * Substitution block
		 *
		 * Try to match the mark with some metadata
		 * in the stream
		 */

		/* Search in the table for key */
		// g_printf("is key '%c', ", fmt[i+1]);
		for ( j = 0; (tag_fmt_table[j].key) != 0 && !key_found; j++ ) {
			if ( tag_fmt_table[j].key == fmt[i+1] ) {
				tag = (gchar *) eina_stream_get_metadata(self, tag_fmt_table[j].tag);
				if ( tag  != NULL ) {
					// g_printf(" [tag (%s) value (%s)] ", eina_tag_get_string_from_tag(tag_fmt_table[j].tag), tag);
					key_found = TRUE;
					g_string_append_safe(string, tag);
				}
			}
		}

		/* If substition was ok, go to next char */
		if ( key_found ) {
			// g_printf("key found (%s)\n", tag);
			i++;
			key_found = FALSE;
			continue;
		} /* else {
			g_printf("key NOT found\n");
		} */
 
		/* Error in substitution, apply fallback methods */
		string = g_string_append_c(string, fmt[i]);
		max_fails--;
		// g_printf("max fails now %d\n", max_fails);
		/* If we didnt reached max errors we can continue processing */
		if ( max_fails >= 0 ) 
			continue;
		
		// g_printf("max fails reached\n");
		/*
		 * Fallback block
		 * 
		 * At this point, we have to discart previous processing
		 * and use uri, stream or filename as base
		 */
		if ( flags & EINA_STREAM_LOCATION )
			string = g_string_assign(string, self->stream);
		else if ( flags & EINA_STREAM_URI )
			string = g_string_assign(string, self->uri);
		else
			/* 
			 * We should handle EINA_STREAM_FILENAME here
			 * This implies check if file is local and 
			 * call to g_filename_from_uri
			 */
			string = g_string_assign(string, self->stream);
	
		/* Stop processing format string */
		break;
	}

	/* Drop the GString */
	ret = string->str;
	g_string_free(string, FALSE);
	
	/* 
	 * Apply flags like dirname or basename 
	 */
	if ( flags & EINA_STREAM_BASENAME ) {
		tmp = g_path_get_basename(ret);
		g_free(ret);
		ret = tmp;
	}
	if ( flags & EINA_STREAM_DIRNAME ) {
		tmp = g_path_get_dirname(ret);
		g_free(ret);
		ret = tmp;
	}
	/* Nothing to do with fullname */

	/* 
	 * Apply flags like utf8 and escape
	 */
	if ( flags & EINA_STREAM_URL_DECODE ) {
		tmp = url_decode(ret);
		g_free(ret);
		ret = tmp;
	}
	
	if ( flags & EINA_STREAM_URL_DECODE ) {
		tmp = url_decode(ret);
		g_free(ret);
		ret = tmp;
	}
	
	if ( flags & EINA_STREAM_ESCAPE ) {
		tmp = g_markup_escape_text(ret, -1);
		g_free(ret);
		ret = tmp;
	}

	if ( flags & EINA_STREAM_UTF8 ) {
		if ( !g_utf8_validate(ret, -1, NULL)) {
			g_warning("String '%s' NOT in utf8\n", ret);
		}
	}
	*dest = ret;

	if ( max_fails >= 0 )
		return TRUE;
	else
		return FALSE;
}

void eina_stream_print(const EinaStream *self) {
	g_printf("[libeina::stream] %p '%s'\n", self, self->uri);
}

gchar *eina_steam_serialize(EinaStream *self) {
	gchar *local;
	gchar *str;

	local = g_strdup_printf("%d", self->is_local);

	str = g_strjoin(EINA_STREAM_SERIALIZE_CHAR,
			self->stream,
			self->uri,
			self->protocol,
			local,
			self->basename);
			/* XXX metadata */
			// self->title, self->artist, self->album);

	g_free(local);
	return str;
}

EinaStream *eina_stream_unserialize(const gchar *str) {
	EinaStream *self;
	gchar **fields;
	
	fields = g_strsplit(str, EINA_STREAM_SERIALIZE_CHAR, 0);
	self = g_new0(EinaStream, 1);
	self->stream   = g_strdup(fields[0]);
	self->uri      = g_strdup(fields[1]);
	self->protocol = g_strdup(fields[2]);
	self->is_local = g_ascii_strtod(fields[3], NULL) / 1;
	self->basename = g_strdup(fields[4]);
	/* XXX: Metadata */
	
	return self;
}

gchar *url_decode(gchar *str) {
	GString *out;
	gint i;
	gint in_len;
	gchar *out_str;
	gboolean hit = FALSE;

	gint a;
	
	in_len = strlen(str);
	out = g_string_new_len(NULL, in_len);

	i = 0;
	while ( i < (in_len - 2) ) {
		if ( str[i] == '%'
			&& g_ascii_isxdigit(str[i+1])
			&& g_ascii_isxdigit(str[i+2])) {
			a = (__hex2int(str[i+1]) * 16) + __hex2int(str[i+2]);
			out = g_string_append_c(out, a);
			hit = TRUE;
		} else {
			out = g_string_append_c(out, str[i]);
			hit = FALSE;
		}

		if ( hit )
			i += 3;
		else
			i += 1;
	}

	if ( !hit ) {
		out = g_string_append_c(out, str[in_len - 2]);
		out = g_string_append_c(out, str[in_len - 1]);
	}
	
	out_str = out->str;
	g_string_free(out, FALSE);

	return out_str;
}

gint __hex2int(gchar hex) {
	if ( ('0' <= hex) && (hex <= '9') )
		return hex - '0';

	if ( ('A' <= hex) && (hex <= 'F') )
		return hex - 'A' + 10;

	if ( ('a' <= hex) && (hex <= 'f') )
		return hex - 'a' + 10;

	return 0;
}
